/*
Copyright (C) 2002-2003 Victor Luchits

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// cg_public.h -- client game dll information visible to engine

#define	CGAME_API_VERSION	21

//
// structs and variables shared with the main engine
//

#define	MAX_PARSE_ENTITIES	1024

typedef struct
{
	qboolean		valid;			// cleared if delta parsing was invalid
	int				serverFrame;
	unsigned int	serverTime;		// time in the server when frame was created
	qboolean		delta;
	qboolean		allentities;
	qboolean		multipov;
	int				deltaFrameNum;
	qbyte			areabits[MAX_CM_AREAS/8];		// portalarea visibility bits
	int				numplayers;
	player_state_t	playerState;
	player_state_t	playerStates[MAX_CLIENTS];
	int				numEntities;
	entity_state_t	parsedEntities[MAX_PARSE_ENTITIES];
	match_state_t	match;

} frame_t;

//===============================================================

//
// functions provided by the main engine
//
typedef struct
{
	// drops to console a client game error
	void			(*Error)( char *msg );

	// console messages
	void			(*Print)( char *msg );

	// dynvars
	dynvar_t	*(*Dynvar_Create)( const char *name, qboolean console, dynvar_getter_f getter, dynvar_setter_f setter );
	void		(*Dynvar_Destroy)( dynvar_t *dynvar );
	dynvar_t	*(*Dynvar_Lookup)( const char *name );
	const char	*(*Dynvar_GetName)( dynvar_t *dynvar );
	dynvar_get_status_t (*Dynvar_GetValue)( dynvar_t *dynvar,void **value );
	dynvar_set_status_t (*Dynvar_SetValue)( dynvar_t *dynvar,void *value );
	void		(*Dynvar_AddListener)( dynvar_t *dynvar, dynvar_listener_f listener );
	void		(*Dynvar_RemoveListener)( dynvar_t *dynvar, dynvar_listener_f listener );

	// console variable interaction
	cvar_t			*(*Cvar_Get)( const char *name, const char *value, int flags );
	cvar_t			*(*Cvar_Set)( const char *name, const char *value );
	void			(*Cvar_SetValue)( const char *name, float value );
	cvar_t			*(*Cvar_ForceSet)( const char *name, const char *value );	// will return 0 0 if not found
	float			(*Cvar_VariableValue)( const char *name );
	char			*(*Cvar_VariableString)( const char *name );

	int				(*Cmd_Argc)( void );
	char			*(*Cmd_Argv)( int arg );
	char			*(*Cmd_Args)( void );	// concatenation of all argv >= 1

	void			(*Cmd_AddCommand)( char *name, void(*cmd)( void ) );
	void			(*Cmd_RemoveCommand)( char *cmd_name );
	void			(*Cmd_ExecuteText)( int exec_when, char *text );
	void			(*Cmd_Execute)( void );

	// files will be memory mapped read only
	// the returned buffer may be part of a larger pak file,
	// or a discrete file from anywhere in the quake search path
	// a -1 return means the file does not exist
	// NULL can be passed for buf to just determine existance
	int				(*FS_FOpenFile)( const char *filename, int *filenum, int mode );
	int				(*FS_Read)( void *buffer, size_t len, int file );
	int				(*FS_Write)( const void *buffer, size_t len, int file );
	int				(*FS_vPrintf)( int file, const char *fmt, va_list argptr ); // wsw -- PLX
	int				(*FS_Tell)( int file );
	int				(*FS_Seek)( int file, int offset, int whence );
	int				(*FS_Eof)( int file );
	int				(*FS_Flush)( int file );
	void			(*FS_FCloseFile)( int file );
	int				(*FS_GetFileList)( const char *dir, const char *extension, char *buf, size_t bufsize );
	const char		*(*FS_FirstExtension)( const char *filename, char *extensions );
	qboolean		(*FS_IsPureFile)( const char *filename );

	// key bindings
	char			*(*Key_GetBindingBuf)( int binding );
	char			*(*Key_KeynumToString)( int keynum );

	void			(*GetConfigString)( int i, char *str, int size );
	unsigned int	(*Milliseconds)( void );
	qboolean		(*DownloadRequest)( const char *filename, qboolean requestpak );

	void			(*NET_GetUserCmd)( int frame, usercmd_t *cmd );
	int				(*NET_GetCurrentUserCmdNum)( void );
	void			(*NET_GetCurrentState)( int *incomingAcknowledged, int *outgoingSequence );

	// refresh system
	void			(*R_UpdateScreen)( void );
	int				(*R_GetClippedFragments)( vec3_t origin, float radius, vec3_t axis[3], int maxfverts, vec3_t *fverts, int maxfragments, fragment_t *fragments );
	void			(*R_ClearScene)( void );
	void			(*R_AddEntityToScene)( entity_t *ent );
	void			(*R_AddLightToScene)( vec3_t org, float intensity, float r, float g, float b, struct shader_s *shader );
	
	void			(*R_AddPolyToScene)( struct shader_s *hShader, int numVerts, struct polyVert_s *verts );
	
	void			(*R_AddLightStyleToScene)( int style, float r, float g, float b );
	void			(*R_RenderScene)( refdef_t *fd );
	void			(*R_RegisterWorldModel)( char *name );
	void			(*R_ModelBounds)( struct model_s *mod, vec3_t mins, vec3_t maxs );
	struct model_s		*(*R_RegisterModel)( char *name );
	struct shader_s 	*(*R_RegisterPic)( char *name );
	struct shader_s 	*(*R_RegisterSkin)( char *name );
	struct skinfile_s 	*(*R_RegisterSkinFile)( char *name );
	qboolean		(*R_LerpTag)( orientation_t *orient, struct model_s *mod, int oldframe, int frame, float lerpfrac, const char *name );
#ifdef CGAMEGETLIGHTORIGIN
	//void R_LightGridForOrigin( vec3_t origin, vec3_t dir, vec4_t ambient, vec4_t diffuse, float radius )
	void			(*R_LightGridForOrigin)( vec3_t origin, vec3_t dir, vec4_t ambient, vec4_t diffuse, float radius );
#endif
//wsw:jal:MAPTEAMCOLORS
	void			(*R_SetTeamColor)( int team, int r, int g, int b );
//wsw:jal:MAPTEAMCOLORS
	void			(*R_DrawStretchPic)( int x, int y, int w, int h, float s1, float t1, float s2, float t2, vec4_t color, struct shader_s *shader );
	void			(*R_TransformVectorToScreen)( refdef_t *rd, vec3_t in, vec2_t out );
	int				(*R_SkeletalGetNumBones)( struct model_s *mod, int *numFrames );
	int				(*R_SkeletalGetBoneInfo)( struct model_s *mod, int bone, char *name, int size, int *flags );
	void			(*R_SkeletalGetBonePose)( struct model_s *mod, int bone, int frame, bonepose_t *bonepose );

	// collision detection
	int				(*CM_NumInlineModels)( void );
	struct cmodel_s	*(*CM_InlineModel)( int num );
	struct cmodel_s	*(*CM_ModelForBBox)( vec3_t mins, vec3_t maxs );
	void			(*CM_BoxTrace)( trace_t *tr, vec3_t start, vec3_t end, vec3_t mins, vec3_t maxs, struct cmodel_s *cmodel, int brushmask );
	void			(*CM_TransformedBoxTrace)( trace_t *tr, vec3_t start, vec3_t end, vec3_t mins, vec3_t maxs, struct cmodel_s *cmodel, int brushmask, vec3_t origin, vec3_t angles );
	int				(*CM_PointContents)( vec3_t p, struct cmodel_s *cmodel );
	int				(*CM_TransformedPointContents)( vec3_t p, struct cmodel_s *cmodel, vec3_t origin, vec3_t angles );
	void			(*CM_InlineModelBounds)( struct cmodel_s *cmodel, vec3_t mins, vec3_t maxs );

	// sound system
	struct sfx_s	*(*S_RegisterSound)( const char *name );
	void			(*S_StartFixedSound)( struct sfx_s *sfx, const vec3_t origin, int entchannel, float fvol, float attenuation );
	void			(*S_StartRelativeSound)( struct sfx_s *sfx, int entnum, int entchannel, float fvol, float attenuation );
	void			(*S_StartGlobalSound)( struct sfx_s *sfx, int entchannel, float fvol );
	void			(*S_Update)( const vec3_t origin, const vec3_t velocity, const vec3_t v_forward, const vec3_t v_right, const vec3_t v_up );
	void			(*S_AddLoopSound)( struct sfx_s *sfx, int entnum, float fvol, float attenuation );
	void			(*S_StartBackgroundTrack)( const char *intro, const char *loop );
	void			(*S_StopBackgroundTrack)( void );

	// fonts
	struct mufont_s *(*SCR_RegisterFont)( char *name );
	void			(*SCR_DrawString)( int x, int y, int align, const char *str, struct mufont_s *font, vec4_t color );
	int				(*SCR_DrawStringWidth)( int x, int y, int align, const char *str, int maxwidth, struct mufont_s *font, vec4_t color );
	size_t			(*SCR_strHeight)( struct mufont_s *font );
	size_t			(*SCR_strWidth)( const char *str, struct mufont_s *font, int maxlen );
	size_t			(*SCR_StrlenForWidth)( const char *str, struct mufont_s *font, size_t maxwidth );

	// managed memory allocation
	struct mempool_s *(*Mem_AllocPool)( const char *name, const char *filename, int fileline );
	void			*(*Mem_Alloc)( struct mempool_s *pool, size_t size, const char *filename, int fileline );
	void			(*Mem_Free)( void *data, const char *filename, int fileline );
	void			(*Mem_FreePool)( struct mempool_s **pool, const char *filename, int fileline );
	void			(*Mem_EmptyPool)( struct mempool_s *pool, const char *filename, int fileline );
} cgame_import_t;

//
// functions exported by the client game subsystem
//
typedef struct
{
	// if API is different, the dll cannot be used
	int				(*API)( void );

	// the init function will be called at each restart
	void			(*Init)( int playerNum, int vidWidth, int vidHeight, qboolean demoplaying, qboolean pure );

	void			(*Shutdown)( void );

	void			(*GameCommand)( int target );
	void			(*ConfigString)( int number, char *value );

	void			(*InGameMenu)( void );

	void			(*LoadLayout)( char *s );

	void			(*NewFrameSnap)( frame_t *frame, frame_t *oldframe );

	void			(*GlobalSound)( vec3_t origin, int entnum, int entChannel, int soundNum, float fvol, float attenuation );
	void			(*GetEntitySpatilization)( int entNum, vec3_t origin, vec3_t velocity );

	void			(*Trace)( trace_t *tr, vec3_t start, vec3_t mins, vec3_t maxs, vec3_t end, int passent, int contentmask );
	void			(*RenderView)( float frameTime, int realTime, unsigned int serverTime, float stereo_separation );

} cgame_export_t;
