/*
Copyright (C) 1997-2001 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "server.h"

/*
==============
SV_DemoRecordStartMessage
==============
*/
static void SV_DemoRecordStartMessage( void )
{
	qbyte	buf_data[MAX_MSGLEN];
	msg_t	buf;
	unsigned int i;
	int		len, numpure;
	entity_state_t	nullstate;
	entity_state_t	*base;

	assert( svs.demofile );
	assert( svs.demofilename );

	//
	// write a single giant fake message with all the startup info
	//
	MSG_Init( &buf, buf_data, sizeof(buf_data) );

	//
	// serverdata needs to go over for all types of servers
	// to make sure the protocol is right, and to set the gamedir
	//
	// send the serverdata
	MSG_WriteByte( &buf, svc_serverdata );
	MSG_WriteLong( &buf, PROTOCOL_VERSION );
	MSG_WriteLong( &buf, svs.spawncount );
	MSG_WriteString( &buf, FS_BaseGameDirectory() );
	MSG_WriteString( &buf, FS_GameDirectory() );
	MSG_WriteShort( &buf, -1 );
	// send full levelname
	MSG_WriteString( &buf, sv.name );
	MSG_WriteByte( &buf, 0 ); // sv_bitflags

	// purelist
	i = 0;
	numpure = 0;
	while( i < svs.purelist_size && strlen(svs.purelist + i) )
	{
		i += strlen( svs.purelist + i );
		i++;
		numpure++;
	}
	if( numpure > (short)0x7fff )
		Com_Error( ERR_DROP, "Error: Too many pure files." );

	MSG_WriteShort( &buf, numpure );

	i = 0;
	while( i < svs.purelist_size && strlen(svs.purelist + i) )
	{
		MSG_WriteString( &buf, svs.purelist + i );
		MSG_WriteLong( &buf, FS_PakChecksum(svs.purelist + i) );
		i += strlen( svs.purelist + i );
		i++;
	}

	// write it to the demo file
	len = LittleLong( buf.cursize );
	FS_Write( &len, 4, svs.demofile );
	FS_Write( buf.data, buf.cursize, svs.demofile );
	MSG_Clear( &buf );

	// config strings
	for( i = 0; i < MAX_CONFIGSTRINGS; i++ )
	{
		if( sv.configstrings[i][0] )
		{
			if( buf.cursize + strlen (sv.configstrings[i]) + 64 > (unsigned int)buf.maxsize )
			{	// write it out
				len = LittleLong( buf.cursize );
				FS_Write( &len, 4, svs.demofile );
				FS_Write( buf.data, buf.cursize, svs.demofile );
				MSG_Clear( &buf );
			}

			MSG_WriteByte( &buf, svc_servercs );
			MSG_WriteString( &buf, va("cs %i \"%s\"", i, sv.configstrings[i]) );
		}
	}

	// baselines
	memset( &nullstate, 0, sizeof(nullstate) );

	for( i = 0; i < MAX_EDICTS; i++ )
	{
		base = &sv.baselines[i];
		if( base->modelindex || base->sound || base->effects )
		{
			if( buf.cursize + 128 > (unsigned int)buf.maxsize )
			{	// write it out
				len = LittleLong( buf.cursize );
				FS_Write( &len, 4, svs.demofile );
				FS_Write( buf.data, buf.cursize, svs.demofile );
				MSG_Clear( &buf );
			}

			MSG_WriteByte( &buf, svc_spawnbaseline );
#ifdef COLLISION4D
			MSG_WriteDeltaEntity( &nullstate, base, &buf, qtrue, qtrue, vec3_origin );
#else
			MSG_WriteDeltaEntity( &nullstate, base, &buf, qtrue, qtrue );
#endif
		}
	}

	MSG_WriteByte( &buf, svc_servercs );
	MSG_WriteString( &buf, "precache" );

	len = LittleLong( buf.cursize );
	FS_Write( &len, 4, svs.demofile );
	FS_Write( buf.data, buf.cursize, svs.demofile );
	MSG_Clear( &buf );

	memset( &svs.democlient, 0, sizeof(svs.democlient) );

	svs.democlient.wholelevel = qtrue;

	svs.democlient.reliableAcknowledge = 0;
	svs.democlient.reliableSequence = 0;
	svs.democlient.reliableSent = 0;
	memset( &svs.democlient.reliableCommands, 0, sizeof(svs.democlient.reliableCommands) );

	svs.democlient.lastframe = sv.framenum - 1;
	svs.democlient.nodelta = qtrue;

	MSG_Init( &svs.democlient.soundsmsg, svs.democlient.soundsmsgData, sizeof(svs.democlient.soundsmsgData) );
	svs.democlient.soundsmsg.allowoverflow = qtrue;
}

/*
==============
SV_DemoRecord_f

Begins server demo recording.  Every entity and every message will be
recorded, but no playerinfo will be stored.  Primarily for demo merging.
==============
*/
void SV_DemoRecord_f (void)
{
	int		demofilename_size, i;

	if( Cmd_Argc() != 2 ) {
		Com_Printf( "Usage: serverrecord <demoname>\n" );
		return;
	}

	if( svs.demofile ) {
		Com_Printf( "Already recording\n" );
		return;
	}

	if( sv.state != ss_game ) {
		Com_Printf( "Must be in a level to record\n" );
		return;
	}

	for( i = 0; i < sv_maxclients->integer; i++ ) {
		if( svs.clients[i].state >= CS_SPAWNED && svs.clients[i].edict &&
			!(svs.clients[i].edict->r.svflags & SVF_NOCLIENT) )
			break;
	}
	if( i == sv_maxclients->integer ) {
		Com_Printf( "No players in game, can't record a demo\n" );
		return;
	}

	//
	// open the demo file
	//

	// real name
	demofilename_size =
		sizeof(char) * (strlen("demos/server/") + strlen(Cmd_Argv(1)) + strlen(va(".wd%d", PROTOCOL_VERSION)) + 1);
	svs.demofilename = Mem_ZoneMalloc( demofilename_size );

	Q_snprintfz( svs.demofilename, demofilename_size, "demos/server/%s", Cmd_Argv(1) );
	COM_UserFilename( svs.demofilename );

	if( !COM_ValidateRelativeFilename(svs.demofilename) ) {
		Mem_ZoneFree( svs.demofilename );
		svs.demofilename = NULL;
		Com_Printf( "Invalid filename.\n" );
		return;
	}

	COM_DefaultExtension( svs.demofilename, va(".wd%d", PROTOCOL_VERSION), demofilename_size );

	// temp name
	demofilename_size = sizeof(char) * (strlen(svs.demofilename) + strlen(".rec") + 1 );
	svs.demotempname = Mem_ZoneMalloc( demofilename_size );
	Q_snprintfz( svs.demotempname, demofilename_size, "%s.rec", svs.demofilename );

	// open it
	if( FS_FOpenFile(svs.demotempname, &svs.demofile, FS_WRITE) == -1 )
	{
		Com_Printf( "ERROR: Couldn't open file: %s\n", svs.demotempname );
		Mem_ZoneFree( svs.demofilename );
		svs.demofilename = NULL;
		Mem_ZoneFree( svs.demotempname );
		svs.demotempname = NULL;
		return;
	}

	Com_Printf( "Recording server demo: %s\n", svs.demofilename );

	SV_DemoRecordStartMessage();
}

/*
==============
SV_DemoRecordStop_f

Console command for stopping server demo recording.
==============
*/
void SV_DemoRecordStop_f( void )
{
	if( !svs.demofile ) {
		Com_Printf( "No server demo recording in progress\n" );
		return;
	}

	Com_Printf( "Stopped server demo recording: %s\n", svs.demofilename );

	FS_FCloseFile( svs.demofile );
	svs.demofile = 0;

	if( !FS_MoveFile(svs.demotempname, svs.demofilename) ) {
		Com_Printf( "Error: Failed to rename the server demo file\n" );
		return;
	}

	Mem_ZoneFree( svs.demofilename );
	svs.demofilename = NULL;
	Mem_ZoneFree( svs.demotempname );
	svs.demotempname = NULL;
}

/*
==============
SV_DemoRecordCancel_f

Cancels the server demo recording (stop, remove file)
==============
*/
void SV_DemoRecordCancel_f( void )
{
	if( !svs.demofile ) {
		Com_Printf( "No server demo recording in progress\n" );
		return;
	}

	Com_Printf( "Canceled server demo recording: %s\n", svs.demofilename );

	FS_FCloseFile( svs.demofile );
	svs.demofile = 0;

	if( !FS_RemoveFile(svs.demotempname) ) {
		Com_Printf( "Error: Failed to delete the temporary server demo file\n" );
		return;
	}

	Mem_ZoneFree( svs.demofilename );
	svs.demofilename = NULL;
	Mem_ZoneFree( svs.demotempname );
	svs.demotempname = NULL;
}

/*
==================
SV_DemoRecordMessage

Save everything in the world out without deltas.
Used for recording footage for merged or assembled demos
==================
*/
void SV_DemoRecordMessage (void)
{
	int				i, len;
	unsigned int	ui;
	qbyte			msg_buf[MAX_MSGLEN];
	msg_t			msg;

	if( !svs.demofile )
		return;

	for( i = 0; i < sv_maxclients->integer; i++ ) {
		if( svs.clients[i].state >= CS_SPAWNED && svs.clients[i].edict &&
			!(svs.clients[i].edict->r.svflags & SVF_NOCLIENT) )
			break;
	}
	if( i == sv_maxclients->integer ) {
		Com_Printf( "No players left, stopping server side demo recording.\n" );
		SV_DemoRecordStop_f();
		return; // skip frames without players
	}

	MSG_Init( &msg, msg_buf, sizeof(msg_buf) );
	MSG_Clear( &msg );
	msg.allowoverflow = qfalse;

	// send over all the relevant entity_state_t
	// and the player_state_t
	SV_BuildClientFrameSnap( &svs.democlient );
	SV_WriteFrameSnapToClient( &svs.democlient, &msg );
	SV_WriteClientSoundsDatagram( &svs.democlient, &msg, &svs.democlient.soundsmsg );
	svs.democlient.lastframe = sv.framenum;

	// write any unsent game commands
	// can't use general function, because of svc_servercmd
	for( ui = svs.democlient.reliableAcknowledge + 1 ; ui <= svs.democlient.reliableSequence ; ui++ ) {
		if( !strlen( svs.democlient.reliableCommands[ ui & (MAX_RELIABLE_COMMANDS-1) ] ) )
			continue;
		MSG_WriteByte( &msg, svc_servercmd );
		MSG_WriteLong( &msg, ui );
		MSG_WriteString( &msg, svs.democlient.reliableCommands[ ui & (MAX_RELIABLE_COMMANDS-1) ] );
	}
	svs.democlient.reliableSent = svs.democlient.reliableSequence;
	svs.democlient.reliableAcknowledge = svs.democlient.reliableSent;

	// now write the entire message to the file, prefixed by the length
	len = LittleLong( msg.cursize );
	FS_Write( &len, 4, svs.demofile );
	FS_Write( msg.data, msg.cursize, svs.demofile );
}
